@@ -28,6 +28,7 @@ typedef struct {
 
 static njs_int_t njs_externals_262_init(njs_vm_t *vm);
 static njs_int_t njs_externals_shared_preinit(njs_vm_t *vm);
+static njs_int_t njs_externals_shared_init(njs_vm_t *vm);
 njs_int_t njs_array_buffer_detach(njs_vm_t *vm, njs_value_t *args,
     njs_uint_t nargs, njs_index_t unused, njs_value_t *retval);
 
@@ -47,7 +48,7 @@ njs_module_t  njs_unit_test_262_module = {
 njs_module_t  njs_unit_test_external_module = {
     .name = njs_str("external"),
     .preinit = njs_externals_shared_preinit,
-    .init = NULL,
+    .init = njs_externals_shared_init,
 };
 
 
@@ -415,7 +416,6 @@ njs_unit_test_r_subrequest(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 {
     njs_int_t            ret;
     njs_value_t          value, *argument, *select;
-    njs_vm_event_t       vm_event;
     njs_function_t       *callback;
     njs_external_ev_t    *ev;
     njs_external_env_t   *env;
@@ -444,16 +444,10 @@ njs_unit_test_r_subrequest(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         return NJS_ERROR;
     }
 
-    vm_event = njs_vm_add_event(vm, callback, 1, NULL, NULL);
-    if (vm_event == NULL) {
-        njs_internal_error(vm, "njs_vm_add_event() failed");
-        return NJS_ERROR;
-    }
-
     argument = njs_arg(args, nargs, 1);
     select = njs_arg(args, nargs, 2);
 
-    ev->vm_event = vm_event;
+    ev->function = callback;
     ev->data = r;
     ev->nargs = 2;
     njs_value_assign(&ev->args[0], &ev->callbacks[!!njs_bool(select)]);
@@ -1260,77 +1254,15 @@ njs_externals_init_internal(njs_vm_t *vm, njs_unit_test_req_init_t *init,
 {
     njs_int_t             ret;
     njs_uint_t            i, j;
-    njs_function_t        *f;
-    njs_opaque_value_t    value;
     njs_unit_test_req_t   *requests;
     njs_unit_test_prop_t  *prop;
 
-    static const njs_str_t  external_ctor = njs_str("ExternalConstructor");
-    static const njs_str_t  external_null = njs_str("ExternalNull");
-    static const njs_str_t  external_error = njs_str("ExternalError");
-
-    if (shared) {
-        njs_external_r_proto_id = njs_vm_external_prototype(vm,
-                                         njs_unit_test_r_external,
-                                         njs_nitems(njs_unit_test_r_external));
-        if (njs_slow_path(njs_external_r_proto_id < 0)) {
-            njs_printf("njs_vm_external_prototype() failed\n");
-            return NJS_ERROR;
-        }
-
-        f = njs_vm_function_alloc(vm, njs_unit_test_constructor, 1, 1);
-        if (f == NULL) {
-            njs_printf("njs_vm_function_alloc() failed\n");
-            return NJS_ERROR;
-        }
-
-        njs_value_function_set(njs_value_arg(&value), f);
-
-        ret = njs_vm_bind(vm, &external_ctor, njs_value_arg(&value), 1);
-        if (njs_slow_path(ret != NJS_OK)) {
-            njs_printf("njs_vm_bind() failed\n");
-            return NJS_ERROR;
-        }
-
-        njs_external_null_proto_id = njs_vm_external_prototype(vm,
-                                       njs_unit_test_null_external,
-                                       njs_nitems(njs_unit_test_null_external));
-        if (njs_slow_path(njs_external_null_proto_id < 0)) {
-            njs_printf("njs_vm_external_prototype() failed\n");
-            return NJS_ERROR;
-        }
-
-        ret = njs_vm_external_create(vm, njs_value_arg(&value),
-                                     njs_external_null_proto_id, NULL, 1);
-        if (njs_slow_path(ret != NJS_OK)) {
-            return NJS_ERROR;
-        }
-
-        ret = njs_vm_bind(vm, &external_null, njs_value_arg(&value), 1);
-        if (njs_slow_path(ret != NJS_OK)) {
-            njs_printf("njs_vm_bind() failed\n");
-            return NJS_ERROR;
-        }
-
-        njs_external_error_ctor_id =
-            njs_vm_external_constructor(vm, &external_error,
-                              njs_error_constructor, njs_unit_test_ctor_props,
-                              njs_nitems(njs_unit_test_ctor_props),
-                              njs_unit_test_proto_props,
-                              njs_nitems(njs_unit_test_proto_props));
-        if (njs_slow_path(njs_external_error_ctor_id < 0)) {
-            njs_printf("njs_vm_external_constructor() failed\n");
-            return NJS_ERROR;
-        }
-    }
-
     requests = njs_mp_zalloc(vm->mem_pool, n * sizeof(njs_unit_test_req_t));
     if (njs_slow_path(requests == NULL)) {
         return NJS_ERROR;
     }
 
     for (i = 0; i < n; i++) {
-
         requests[i] = init[i].request;
 
         ret = njs_vm_external_create(vm, njs_value_arg(&requests[i].value),
@@ -1403,7 +1335,76 @@ njs_externals_262_init(njs_vm_t *vm)
 static njs_int_t
 njs_externals_shared_preinit(njs_vm_t *vm)
 {
-    return njs_externals_init_internal(vm, njs_test_requests, 1, 1);
+    static const njs_str_t  external_error = njs_str("ExternalError");
+
+    njs_external_r_proto_id = njs_vm_external_prototype(vm,
+                                     njs_unit_test_r_external,
+                                     njs_nitems(njs_unit_test_r_external));
+    if (njs_slow_path(njs_external_r_proto_id < 0)) {
+        njs_printf("njs_vm_external_prototype() failed\n");
+        return NJS_ERROR;
+    }
+
+    njs_external_null_proto_id = njs_vm_external_prototype(vm,
+                                   njs_unit_test_null_external,
+                                   njs_nitems(njs_unit_test_null_external));
+    if (njs_slow_path(njs_external_null_proto_id < 0)) {
+        njs_printf("njs_vm_external_prototype() failed\n");
+        return NJS_ERROR;
+    }
+
+    njs_external_error_ctor_id =
+        njs_vm_external_constructor(vm, &external_error,
+                          njs_error_constructor, njs_unit_test_ctor_props,
+                          njs_nitems(njs_unit_test_ctor_props),
+                          njs_unit_test_proto_props,
+                          njs_nitems(njs_unit_test_proto_props));
+    if (njs_slow_path(njs_external_error_ctor_id < 0)) {
+        njs_printf("njs_vm_external_constructor() failed\n");
+        return NJS_ERROR;
+    }
+
+    return NJS_OK;
+}
+
+
+static njs_int_t
+njs_externals_shared_init(njs_vm_t *vm)
+{
+    njs_int_t           ret;
+    njs_function_t      *f;
+    njs_opaque_value_t  value;
+
+    static const njs_str_t  external_ctor = njs_str("ExternalConstructor");
+    static const njs_str_t  external_null = njs_str("ExternalNull");
+
+    f = njs_vm_function_alloc(vm, njs_unit_test_constructor, 1, 1);
+    if (f == NULL) {
+        njs_printf("njs_vm_function_alloc() failed\n");
+        return NJS_ERROR;
+    }
+
+    njs_value_function_set(njs_value_arg(&value), f);
+
+    ret = njs_vm_bind(vm, &external_ctor, njs_value_arg(&value), 1);
+    if (njs_slow_path(ret != NJS_OK)) {
+        njs_printf("njs_vm_bind() failed\n");
+        return NJS_ERROR;
+    }
+
+    ret = njs_vm_external_create(vm, njs_value_arg(&value),
+                                 njs_external_null_proto_id, NULL, 1);
+    if (njs_slow_path(ret != NJS_OK)) {
+        return NJS_ERROR;
+    }
+
+    ret = njs_vm_bind(vm, &external_null, njs_value_arg(&value), 1);
+    if (njs_slow_path(ret != NJS_OK)) {
+        njs_printf("njs_vm_bind() failed\n");
+        return NJS_ERROR;
+    }
+
+    return njs_externals_init_internal(vm, &njs_test_requests[0], 1, 1);
 }
 
 
@@ -1429,10 +1430,15 @@ njs_external_env_init(njs_external_env_t *env)
 njs_int_t
 njs_external_process_events(njs_vm_t *vm, njs_external_env_t *env)
 {
+    njs_int_t          ret;
     njs_queue_t        *events;
     njs_queue_link_t   *link;
     njs_external_ev_t  *ev;
 
+    if (env == NULL) {
+        return NJS_OK;
+    }
+
     events = &env->events;
 
     for ( ;; ) {
@@ -1448,11 +1454,13 @@ njs_external_process_events(njs_vm_t *vm, njs_external_env_t *env)
         ev->link.prev = NULL;
         ev->link.next = NULL;
 
-        njs_vm_post_event(vm, ev->vm_event, njs_value_arg(&ev->args[0]),
-                          ev->nargs);
+        ret = njs_vm_call(vm, ev->function, njs_value_arg(ev->args), ev->nargs);
+        if (ret == NJS_ERROR) {
+            return NJS_ERROR;
+        }
     }
 
-    return NJS_OK;
+    return njs_vm_pending(vm) ? NJS_AGAIN: NJS_OK;
 }
 
 
@@ -1474,5 +1482,16 @@ njs_external_call(njs_vm_t *vm, const njs_str_t *fname, njs_value_t *args,
         return NJS_ERROR;
     }
 
-    return njs_vm_run(vm);
+    for ( ;; ) {
+        ret = njs_vm_execute_pending_job(vm);
+        if (ret <= NJS_OK) {
+            if (ret == NJS_ERROR) {
+                return NJS_ERROR;
+            }
+
+            break;
+        }
+    }
+
+    return NJS_OK;
 }
